import { defineComponent, h } from 'vue'
import previewMixin from './js/previewMixin'
import renderMixin from './js/renderMixin'
import methodMixin from './js/methodMixin'
import eventMixin from './js/eventMixin'
import change from './js/change'
import handlersMixin from './js/handlersMixin'
import hook from './js/hook'
import watch from './js/watch'
import build from './js/build'
import { CLASS_HIDDEN, CLASS_HIDE, MIME_TYPE_JPEG, NAMESPACE, REGEXP_DATA_URL, REGEXP_DATA_URL_JPEG, WINDOW } from './js/constants'
import { addClass, addTimestamp, arrayBufferToDataURL, assign, dataURLToArrayBuffer, isCrossOriginURL, parseOrientation, removeClass, resetAndGetOrientation } from './js/utilities'
export default defineComponent({
  name: 'V3Cropper',
  render() {
    return h('div', [h('img', { src: this.src, id: 'element', style: { display: 'none' }})])
  },
  props: {
    src: {
      type: [String, Blob, null, File],
      default: ''
    },
    /**
     * 定义裁剪器的视图模式
     * 0, 1, 2, 3
     */
    viewMode: {
      type: Number,
      default: 0
    },
    /**
     * 定义裁剪器的拖动模式
     * 'crop', 'move' or 'none'
     */
    dragMode: {
      type: String,
      default: 'crop'
    },
    /**
     * 定义裁剪框的初始纵横比
     */
    initialAspectRatio: {
      type: Number,
      default: NaN
    },
    /**
     * 初始化时，存储的上一个裁剪数据将自动传递给setData方法。
     * 仅当autoCrop选项设置为true时可用。
     */
    data: {
      type: Object,
      default: null
    },
    /**
     * 为预览添加额外的元素（容器）。
     */
    // preview: {
    //   Type: [Element, Object, String],
    //   default: ''
    // },
    /**
     * 调整窗口大小时重新渲染裁剪器。
     */
    responsive: {
      type: Boolean,
      default: true
    },
    /**
     * 调整窗口大小后恢复裁剪区域。
     */
    restore: {
      type: Boolean,
      default: true
    },
    checkCrossOrigin: {
      type: Boolean,
      default: true
    },
    /**
     * 在图像上方和裁剪框下方显示黑色模式。
     */
    modal: {
      type: Boolean,
      default: true
    },
    /**
     * 显示裁剪框上方的虚线。
     */
    guides: {
      type: Boolean,
      default: true
    },
    /**
     * 显示裁剪框上方的中心指示器。
     */
    center: {
      type: Boolean,
      default: true
    },
    /**
     * 在裁剪框上方显示白色模式（高亮显示裁剪框）。
     */
    highlight: {
      type: Boolean,
      default: true
    },
    /**
     * 显示容器的网格背景。
     */
    background: {
      type: Boolean,
      default: true
    },
    /**
     * 启用在初始化时自动裁剪图像。
     */
    autoCrop: {
      type: Boolean,
      default: true
    },
    /**
     * 它应该是一个介于0和1之间的数字。定义自动裁剪区域大小（百分比）。
     */
    autoCropArea: {
      type: Number,
      default: 0.8
    },
    /**
     * 启用以移动图像。
     */
    movable: {
      type: Boolean,
      default: true
    },
    /**
     * 启用以旋转图像。
     */
    rotatable: {
      type: Boolean,
      default: true
    },
    /**
     * 启用缩放图像。
     */
    scalable: {
      type: Boolean,
      default: true
    },
    /**
     * 启用以缩放图像。
     */
    zoomable: {
      type: Boolean,
      default: true
    },
    /**
     * 启用通过拖动触摸缩放图像。
     */
    zoomOnTouch: {
      type: Boolean,
      default: true
    },
    /**
     * 启用通过鼠标滚轮缩放图像。
     */
    zoomOnWheel: {
      type: Boolean,
      default: true
    },
    /**
     * 通过鼠标滚轮缩放图像时定义缩放比例。
     */
    wheelZoomRatio: {
      type: Number,
      default: 0.1
    },
    /**
     * 允许通过拖动移动裁剪框。
     */
    cropBoxMovable: {
      type: Boolean,
      default: true
    },
    /**
     * 允许通过拖动调整裁剪框的大小。
     */
    cropBoxResizable: {
      type: Boolean,
      default: true
    },
    /**
     * 当在裁剪器上单击两次时，启用在“裁剪”和“移动”之间切换拖动模式。
     * 需要dblclick事件支持。
     */
    toggleDragModeOnDblclick: {
      type: Boolean,
      default: true
    },
    /**
     * 容器的最小宽度。
     */
    minContainerWidth: {
      type: Number,
      default: 200
    },
    /**
     * 容器的最小高度。
     */
    minContainerHeight: {
      type: Number,
      default: 100
    },
    /**
     * 画布（图像包装）的最小宽度。
     */
    minCanvasWidth: {
      type: Number,
      default: 0
    },
    /**
     * 画布（图像包装）的最小高度。
     */
    minCanvasHeight: {
      type: Number,
      default: 0
    },
    /**
     * 裁剪框的最小宽度。
     */
    minCropBoxWidth: {
      type: Number,
      default: 0
    },
    /**
     * 裁剪框的最小高度。
     */
    minCropBoxHeight: {
      type: Number,
      default: 0
    },
    /**
     * 可以压缩图片宽高  默认不超过200
     */
    maxImgSize: {
      type: [Number, String],
      default: 2000
    }
  },
  data() {
    return {
      element: {},
      cropped: false,
      disabled: false,
      pointers: {},
      ready: false,
      reloading: false,
      replaced: false,
      sized: false,
      sizing: false,
      checkOrientation: true,
      xhr: {},
      limited: '',
      // image info
      isImg: false,
      originalUrl: '',
      crossOriginUrl: '',
      url: '',
      sizingImage: '',
      // cropper info,
      image: {},
      imageData: {},
      initialImageData: {},
      // container info
      container: {},
      cropper: {},
      canvas: {},
      dragBox: {},
      cropBox: {},
      viewBox: {},
      viewBoxImage: {},
      face: {},
      containerData: {},
      canvasData: {},
      initialCanvasData: {},
      initialCropBoxData: {},
      cropBoxData: {}
    }
  },
  mixins: [previewMixin, renderMixin, methodMixin, eventMixin, change, handlersMixin, hook, watch, build],
  mounted() {
    this.element = document.getElementById('element')
    this.builder()
    this.loadImage()
  },
  methods: {
    loadImage(val) {
      const { element } = this

      element[NAMESPACE] = this

      this.isImg = true

      // e.g.: "img/picture.jpg"
      const url = val || element.getAttribute('src')
      this.originalUrl = url

      // Stop when it's a blank image
      if (!url) {
        return
      }

      // e.g.: "https://example.com/img/picture.jpg"
      this.load(url)
    },
    load(url) {
      if (!url) {
        return
      }
      this.url = url
      this.imageData = {}

      const { element, $props } = this

      if (!$props.rotatable && !$props.scalable) {
        this.checkOrientation = false
      }
      // Only IE10+ supports Typed Arrays
      if (!this.checkOrientation || !window.ArrayBuffer) {
        this.clone()
        return
      }

      // Detect the mime type of the image directly if it is a Data URL
      if (REGEXP_DATA_URL.test(url)) {
        // Read ArrayBuffer from Data URL of JPEG images directly for better performance
        if (REGEXP_DATA_URL_JPEG.test(url)) {
          this.read(dataURLToArrayBuffer(url))
        } else {
          // Only a JPEG image may contains Exif Orientation information,
          // the rest types of Data URLs are not necessary to check orientation at all.
          this.clone()
        }
        return
      }

      // 1. Detect the mime type of the image by a XMLHttpRequest.
      // 2. Load the image as ArrayBuffer for reading orientation if its a JPEG image.
      const xhr = new XMLHttpRequest()
      const clone = this.clone.bind(this)

      this.reloading = true
      this.xhr = xhr

      // 1. Cross origin requests are only supported for protocol schemes:
      // http, https, data, chrome, chrome-extension.
      // 2. Access to XMLHttpRequest from a Data URL will be blocked by CORS policy
      // in some browsers as IE11 and Safari.
      xhr.onabort = clone
      xhr.onerror = clone
      xhr.ontimeout = clone

      xhr.onprogress = () => {
        // Abort the request directly if it not a JPEG image for better performance
        if (xhr.getResponseHeader('content-type') !== MIME_TYPE_JPEG) {
          xhr.abort()
        }
      }

      xhr.onload = () => {
        this.read(xhr.response)
      }

      xhr.onloadend = () => {
        this.reloading = false
        this.xhr = null
      }

      // Bust cache when there is a "crossOrigin" property to avoid browser cache error
      if (this.checkCrossOrigin && isCrossOriginURL(url) && element.crossOrigin) {
        url = addTimestamp(url)
      }

      // The third parameter is required for avoiding side-effect (#682)
      xhr.open('GET', url, true)
      xhr.responseType = 'arraybuffer'
      xhr.withCredentials = element.crossOrigin === 'use-credentials'
      xhr.send()
    },
    read(arrayBuffer) {
      const { $props, imageData } = this

      // Reset the orientation value to its default value 1
      // as some iOS browsers will render image with its orientation
      const orientation = resetAndGetOrientation(arrayBuffer)

      let rotate = 0
      let scaleX = 1
      let scaleY = 1

      if (orientation > 1) {
        // Generate a new URL which has the default orientation value
        this.url = arrayBufferToDataURL(arrayBuffer, MIME_TYPE_JPEG);
        ({ rotate, scaleX, scaleY } = parseOrientation(orientation))
      }

      if ($props.rotatable) {
        imageData.rotate = rotate
      }

      if ($props.scalable) {
        imageData.scaleX = scaleX
        imageData.scaleY = scaleY
      }

      this.clone()
    },
    clone() {
      const { element, url } = this
      let { crossOrigin } = element
      let crossOriginUrl = url
      if (this.checkCrossOrigin && isCrossOriginURL(url)) {
        if (!crossOrigin) {
          crossOrigin = 'anonymous'
        }

        // Bust cache when there is not a "crossOrigin" property (#519)
        crossOriginUrl = addTimestamp(url)
      }

      this.crossOrigin = crossOrigin
      this.crossOriginUrl = crossOriginUrl

      const image = document.createElement('img')
      this.image = image
      if (crossOrigin) {
        image.crossOrigin = crossOrigin
      }

      image.src = crossOriginUrl || url
      image.alt = element.alt || 'The image to crop'
      image.onload = this.start.bind(this)
      image.onerror = this.stop.bind(this)
      addClass(image, CLASS_HIDE)
      element.parentNode.insertBefore(image, element.nextSibling)
    },
    start() {
      return new Promise(resolve => {
        const { image } = this
        image.onload = null
        image.onerror = null
        this.sizing = true

        // Match all browsers that use WebKit as the layout engine in iOS devices,
        // such as Safari for iOS, Chrome for iOS, and in-app browsers.
        const isIOSWebKit = WINDOW.navigator && /(?:iPad|iPhone|iPod).*?AppleWebKit/i.test(WINDOW.navigator.userAgent)

        const done = (naturalWidth, naturalHeight) => {
          assign(this.imageData, {
            naturalWidth,
            naturalHeight,
            aspectRatio: naturalWidth / naturalHeight
          })
          this.initialImageData = assign({}, this.imageData)
          this.sizing = false
          this.sized = true
          this.build()
          resolve()
        }

        // Most modern browsers (excepts iOS WebKit)
        if (image.naturalWidth && !isIOSWebKit) {
          done(image.naturalWidth, image.naturalHeight)
          return
        }

        const sizingImage = document.createElement('img')
        const body = document.body || document.documentElement

        this.sizingImage = sizingImage

        sizingImage.onload = () => {
          done(sizingImage.width, sizingImage.height)

          if (!isIOSWebKit) {
            body.removeChild(sizingImage)
          }
        }

        sizingImage.src = image.src

        // iOS WebKit will convert the image automatically
        // with its orientation once append it into DOM (#279)
        if (!isIOSWebKit) {
          sizingImage.style.cssText = (
            'left:0;' +
            'max-height:none!important;' +
            'max-width:none!important;' +
            'min-height:0!important;' +
            'min-width:0!important;' +
            'opacity:0;' +
            'position:absolute;' +
            'top:0;' +
            'z-index:-1;'
          )
          body.appendChild(sizingImage)
        }
      })
    },
    build() {
      if (!this.sized) {
        return
      }

      const { $props } = this

      this.initPreview()
      this.bind()

      removeClass(this.image, CLASS_HIDDEN)

      this.canvas.appendChild(this.image)

      this.render()
      this.setDragMode()
      if (this.autoCrop) {
        this.crop()
      }

      this.setData($props.data)
    },
    stop() {
      const { image } = this
      image.onload = null
      image.onerror = null
      image.parentNode.removeChild(image)
      this.image = null
    },
    unbuild() {
      if (!this.ready) {
        return
      }

      this.ready = false
      this.unbind()
      this.resetPreview()
      this.cropper.parentNode.removeChild(this.cropper)
      removeClass(this.element, CLASS_HIDDEN)
    },
    uncreate() {
      if (this.ready) {
        this.unbuild()
        this.ready = false
        this.cropped = false
      } else if (this.sizing) {
        this.sizingImage.onload = null
        this.sizing = false
        this.sized = false
      } else if (this.reloading) {
        this.xhr.onabort = null
        this.xhr.abort()
      } else if (this.image) {
        this.stop()
      }
    }
  }
})
